/*
	File:		SerialPortSample.c
	
	Description:	This sample demonstrates how to use IOKitLib to find all serial ports on the system. 
                        It also shows how to open, write to, read from, and close a serial port.
                
	Author:		gc

	Copyright: 	 Copyright 2000 Apple Computer, Inc. All rights reserved.
	
	Disclaimer:	IMPORTANT:  This Apple software is supplied to you by Apple Computer, Inc.
                        ("Apple") in consideration of your agreement to the following terms, and your
                        use, installation, modification or redistribution of this Apple software
                        constitutes acceptance of these terms.  If you do not agree with these terms,
                        please do not use, install, modify or redistribute this Apple software.

                        In consideration of your agreement to abide by the following terms, and subject
                        to these terms, Apple grants you a personal, non-exclusive license, under Apples
                        copyrights in this original Apple software (the "Apple Software"), to use,
                        reproduce, modify and redistribute the Apple Software, with or without
                        modifications, in source and/or binary forms; provided that if you redistribute
                        the Apple Software in its entirety and without modifications, you must retain
                        this notice and the following text and disclaimers in all such redistributions of
                        the Apple Software.  Neither the name, trademarks, service marks or logos of
                        Apple Computer, Inc. may be used to endorse or promote products derived from the
                        Apple Software without specific prior written permission from Apple.  Except as
                        expressly stated in this notice, no other rights or licenses, express or implied,
                        are granted by Apple herein, including but not limited to any patent rights that
                        may be infringed by your derivative works or by other works in which the Apple
                        Software may be incorporated.

                        The Apple Software is provided by Apple on an "AS IS" basis.  APPLE MAKES NO
                        WARRANTIES, EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED
                        WARRANTIES OF NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
                        PURPOSE, REGARDING THE APPLE SOFTWARE OR ITS USE AND OPERATION ALONE OR IN
                        COMBINATION WITH YOUR PRODUCTS.

                        IN NO EVENT SHALL APPLE BE LIABLE FOR ANY SPECIAL, INDIRECT, INCIDENTAL OR
                        CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
                        GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
                        ARISING IN ANY WAY OUT OF THE USE, REPRODUCTION, MODIFICATION AND/OR DISTRIBUTION
                        OF THE APPLE SOFTWARE, HOWEVER CAUSED AND WHETHER UNDER THEORY OF CONTRACT, TORT
                        (INCLUDING NEGLIGENCE), STRICT LIABILITY OR OTHERWISE, EVEN IF APPLE HAS BEEN
                        ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
				
	Change History (most recent first):
        
            <2>	 	08/22/00	gc		Incorporated changes from code review.
            <1>	 	08/03/00	gc		New sample.
        
*/

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <paths.h>
#include <termios.h>
#include <sys/param.h>
#include <IOKit/IOKitLib.h>
#include <IOKit/IOBSD.h>

#define kIOPortDeviceNubClass	"IOPortDeviceNub"
#define kDeviceTypeKey		"device_type"
#define kSerialValue		"serial"
#define kPortNameKey		"Port Name"
#define kModemValue		"modem"
#define kATCommandString	"AT\r"
#define kOKResponseString	"OK"

enum {
    kNumRetries = 3
};

static kern_return_t FindSerialPorts( io_iterator_t *matchingServices );
static kern_return_t FindModem( io_iterator_t serialPortIterator, char *bsdPath, CFIndex maxPathSize );
static int OpenSerialPort( const char *bsdPath );
static Boolean InitializeModem( int fileDescriptor );
static void CloseSerialPort( int fileDescriptor );

// Returns an iterator across all serial ports. Caller is responsible for releasing the iterator
// when iteration is complete.
kern_return_t FindSerialPorts( io_iterator_t *matchingServices )
{
    kern_return_t		kernResult; 
    mach_port_t			masterPort;
    CFMutableDictionaryRef	classesToMatch;

/*! @function IOMasterPort
    @abstract Returns the mach port used to initiate communication with IOKit.
    @discussion Functions that don't specify an existing object require the IOKit master port to be passed. This function obtains that port.
    @param size The task's bootstrap port, pass bootstrap_port from mach/mach_init.h.
    @param masterPort The master port is returned.
    @result A kern_return_t error code. */

    kernResult = IOMasterPort( bootstrap_port, &masterPort );
    if ( KERN_SUCCESS != kernResult )
        printf( "IOMasterPort returned %d\n", kernResult );
        
/*! @function IOServiceMatching
    @abstract Create a matching dictionary that specifies an IOService class match.
    @discussion A very common matching criteria for IOService is based on its class. IOServiceMatching will create a matching dictionary that specifies any IOService of a class, or its subclasses. The class is specified by C-string name.
    @param name The class name, as a const C-string. Class matching is successful on IOService's of this class or any subclass.
    @result The matching dictionary created, is returned on success, or zero on failure. The dictionary is commonly passed to IOServiceGetMatchingServices or IOServiceAddNotification which will consume a reference, otherwise it should be released with CFRelease by the caller. */

    // Serial devices are instances of class kIOPortDeviceNubClass
    classesToMatch = IOServiceMatching( kIOPortDeviceNubClass ); 
    if ( classesToMatch == NULL )
        printf( "IOServiceMatching returned a NULL dictionary.\n" );
    else {
/*!
	@function CFDictionarySetValue
	Sets the value of the key in the dictionary.
	@param theDict The dictionary to which the value is to be set. If this
		parameter is not a valid mutable CFDictionary, the behavior is
		undefined. If the dictionary is a fixed-capacity dictionary and
		it is full before this operation, and the key does not exist in
		the dictionary, the behavior is undefined.
	@param key The key of the value to set into the dictionary. If a key 
		which matches this key is already present in the dictionary, only
		the value is changed ("add if absent, replace if present"). If
		no key matches the given key, the key-value pair is added to the
		dictionary. If added, the key is retained by the dictionary,
		using the retain callback provided
		when the dictionary was created. If the key is not of the sort
		expected by the key retain callback, the behavior is undefined.
	@param value The value to add to or replace into the dictionary. The value
		is retained by the dictionary using the retain callback provided
		when the dictionary was created, and the previous value if any is
		released. If the value is not of the sort expected by the
		retain or release callbacks, the behavior is undefined.
*/

        CFDictionarySetValue( classesToMatch, CFSTR( kDeviceTypeKey ), CFSTR( kSerialValue ) ); 
        // Each serial device object has a property with key kDeviceTypeKey and value kSerialValue
        // so add this property to the CFDictionary we're matching on. 
    }
    
    /*! @function IOServiceGetMatchingServices
        @abstract Look up registered IOService objects that match a matching dictionary.
        @discussion This is the preferred method of finding IOService objects currently registered by IOKit. IOServiceAddNotification can also supply this information and install a notification of new IOServices. The matching information used in the matching dictionary may vary depending on the class of service being looked up.
        @param masterPort The master port obtained from IOMasterPort().
        @param matching A CF dictionary containing matching information, of which one reference is consumed by this function. IOKitLib can contruct matching dictionaries for common criteria with helper functions such as IOServiceMatching, IOOpenFirmwarePathMatching.
        @param existing An iterator handle is returned on success, and should be released by the caller when the iteration is finished.
        @result A kern_return_t error code. */

    kernResult = IOServiceGetMatchingServices( masterPort, classesToMatch, matchingServices );    
    if ( KERN_SUCCESS != kernResult )
        printf( "IOServiceGetMatchingServices returned %d\n", kernResult );
        
    return kernResult;
}
    
// Given an iterator across a set of serial ports, find the first modem and return the BSD path to the
// modem device. If no modems are found the path name is set to an empty string.
kern_return_t FindModem( io_iterator_t serialPortIterator, char *bsdPath, CFIndex maxPathSize )
{
    io_object_t		nextService;
    kern_return_t	kernResult = KERN_FAILURE;
    Boolean		modemFound = false;
    
    /*! @function IOIteratorNext
    @abstract Returns the next object in an iteration.
    @discussion This function returns the next object in an iteration, or zero if no more remain or the iterator is invalid.
    @param iterator An IOKit iterator handle.
    @result If the iterator handle is valid, the next element in the iteration is returned, otherwise zero is returned. */

    while ( nextService = IOIteratorNext( serialPortIterator ) )
    {
        CFTypeRef	portNameAsCFString;
        CFTypeRef	bsdPathAsCFString;
        char		portNameAsCString[ MAXPATHLEN ];
        
/*! @function IORegistryEntryCreateCFProperty
    @abstract Create a CF representation of a registry entry's property.
    @discussion This function creates an instantaneous snapshot of a registry entry property, creating a CF container analogue in the caller's task. Not every object available in the kernel is represented as a CF container; currently OSDictionary, OSArray, OSSet, OSSymbol, OSString, OSData, OSNumber, OSBoolean are created as their CF counterparts. 
    @param entry The registry entry handle whose property to copy.
    @param key A CFString specifying the property name.
    @param allocator The CF allocator to use when creating the CF container.
    @param options No options are currently defined.
    @result A CF container is created and returned the caller on success. The caller should release with CFRelease. */

        portNameAsCFString = IORegistryEntryCreateCFProperty( nextService, 
                                                              CFSTR( kPortNameKey ), 
                                                              kCFAllocatorDefault, 
                                                              0 );

        *portNameAsCString = '\0';
        if ( portNameAsCFString )
        {
            if ( CFStringGetCString( portNameAsCFString, 
                                     portNameAsCString, 
                                     sizeof( portNameAsCString ), 
                                     kCFStringEncodingASCII ) )
            {
                printf( "Port name: %s, ", portNameAsCString );
                kernResult = KERN_SUCCESS;
            }

            CFRelease( portNameAsCFString );
        }

        bsdPathAsCFString = IORegistryEntryCreateCFProperty( nextService, 
                                                             CFSTR( kIOBSDName ), 
                                                             kCFAllocatorDefault, 
                                                             0 );

        *bsdPath = '\0';
        if ( bsdPathAsCFString )
        {
            size_t devPathLength = strlen( _PATH_DEV );
            
            strcpy( bsdPath, _PATH_DEV );
            
            if ( CFStringGetCString( bsdPathAsCFString,
                                     bsdPath + devPathLength,
                                     maxPathSize - devPathLength, 
                                     kCFStringEncodingASCII ) )
            {
                printf( "BSD path: %s", bsdPath );
            }
            
            CFRelease( bsdPathAsCFString );
        }
            
        // See if this port has "modem" in its name
        if ( strstr( portNameAsCString, kModemValue ) )
        {
            modemFound = true;
            printf( " <---modem found!\n" );
            break;
        }
        
        printf( "\n" );

/*! @function IOObjectRelease
    @abstract Releases an object handle previously returned by IOKitLib.
    @discussion All objects returned by IOKitLib should be released with this function when access to them is no longer needed. Using the object after it has been released may or may not return an error, depending on how many references the task has to the same object in the kernel.
    @param object The IOKit object to release.
    @result A kern_return_t error code. */
    
        kernResult = IOObjectRelease( nextService );
    }

    if ( !modemFound )
        *bsdPath = '\0';
        
    return kernResult;
}

// Given the path to a serial device, open the device and configure it.
// Return the file descriptor associated with the device.
int OpenSerialPort( const char *bsdPath )
{
    int 		fileDescriptor;
    struct termios	options;
    
    fileDescriptor = open( bsdPath, O_RDWR | O_NOCTTY | O_NDELAY );
    
    if ( fileDescriptor == -1 )
        printf( "Error %d opening serial port %s.\n", errno, bsdPath );
    else {
        fcntl( fileDescriptor, F_SETFL, 0 );
    
        // Get the current options
        tcgetattr( fileDescriptor, &options );
        
        // Set raw input, one second timeout
        // These options are documented in the man page for termios
        // (in Terminal enter: man termios)
        options.c_cflag |= ( CLOCAL | CREAD );
        options.c_lflag &= ~( ICANON | ECHO | ECHOE | ISIG );
        options.c_oflag &= ~OPOST;
        options.c_cc[ VMIN ] = 0;
        options.c_cc[ VTIME ] = 10;
        
        // Set the options
        tcsetattr( fileDescriptor, TCSANOW, &options );
    }
    
    return fileDescriptor;
}

// Given the file descriptor for a modem device, attempt to initialize the modem by sending it
// a standard AT command and reading the response. If successful the response is "OK".
// Return true if successful, otherwise false.
Boolean InitializeModem( int fileDescriptor )
{
    char	buffer[ 255 ];	// Input buffer
    char	*bufPtr;	// Current char in buffer
    ssize_t	numBytes;	// Number of bytes read or written
    int		tries;		// Number of tries so far
    
    for ( tries = 0; tries < kNumRetries; tries++ )
    {
        // Send an AT command to the modem
        numBytes = write( fileDescriptor, kATCommandString, strlen( kATCommandString ) );
        if ( numBytes < strlen( kATCommandString ) ) 
            continue;
            
        // Read characters into our buffer until we get a CR or LF
        bufPtr = buffer;
        while ( ( numBytes = read( fileDescriptor, bufPtr, buffer + sizeof( buffer ) - bufPtr - 1 ) ) > 0 )
        {
            bufPtr += numBytes;
            if ( *( bufPtr - 1 ) == '\n' || *( bufPtr - 1 ) == '\r' )
                break;
        }
        
        // NUL terminate the string and see if we got an OK response
        *bufPtr = '\0';
        
        if ( strncmp( buffer, kOKResponseString, strlen( kOKResponseString) ) == 0 )
            return false;
    }
    
    return true;
}

// Given the file descriptor for a serial device, close that device.
void CloseSerialPort( int fileDescriptor )
{
    close( fileDescriptor );
}

int main( int argc, char *argv[] )
{
    kern_return_t	kernResult; // on PowerPC this is an int (4 bytes)
/*
 *	error number layout as follows (see mach/error.h):
 *
 *	hi		 		       lo
 *	| system(6) | subsystem(12) | code(14) |
 */

    io_iterator_t	serialPortIterator;
    char		bsdPath[ MAXPATHLEN ];
 
    kernResult = FindSerialPorts( &serialPortIterator );
    
    kernResult = FindModem( serialPortIterator, bsdPath, sizeof( bsdPath ) );
    
    // Now open the modem port we found, initialize the modem, and close the port
    if ( bsdPath[ 0 ] != '\0' )
    {
        int fileDescriptor;
        
        fileDescriptor = OpenSerialPort( bsdPath );
        
        if ( InitializeModem( fileDescriptor ) )
            printf( "Modem initialized.\n" );
        else
            printf( "Could not initialize modem.\n" );
            
        CloseSerialPort( fileDescriptor );
        printf( "Modem port closed.\n" );
    }
    else
        printf( "No modem port found.\n" );
        
    // Release the iterator.
    IOObjectRelease( serialPortIterator );

    return 0;
}